 ////////////////////////////////////////////////////////////////////////////
 //               The TMTCD - Enhanced CD-ROM Audio control unit.          //
 //                                                                        //
 //                    Copyright (C) 1998 Andrew Luzhin                    //
 //                                                                        //
 //  Filename:            SRCfile: tmtcd.pas                               //
 //  Version:             1.41                                             //
 //  Author:              Andrew Luzhin (2:5020/368.31)                    //
 //                       andy21@windoms.sitek.net                         //
 //                       http://windoms.sitek.net/~andy21                 //
 //                                                                        //
 //  Language:            TMT PASCAL                                       //
 //  Supported extenders: PMODE.EXE, PMODE/W and WDOSX compatible          //
 //  Environment:         IBM PC i80386+ (MsDos) 32 bit Protected Mode.    //
 //                       MSCDEX Version 2.10+                             //
 //                                                                        //
 //  Description: This unit provides a number of routines to work with     //
 //               audio CD.                                                //
 //                                                                        //
 //               Don't find it strange that not all functions work        //
 //               (correct) on your CD-ROM... The problems is your device  //
 //               driver (and not this unit, I hope so ;-).                //
 //                                                                        //
 //               If you got any problems or find any bugs please e-mail   //
 //               me. If you use this code it is on your own risk. If you  //
 //               really use this code I would like a copy of your         //
 //               program. Or at least mention my name somewhere...        //
 //                                                                        //
 //  Credits: Yuval Melamed (CDAUDIO unit),Ralf Brown (Interrupt List 58)  //
 //                                                                        //
 //  Modified: 23-Aug-1998, 14:31                                          //
 //                                                                        //
 ////////////////////////////////////////////////////////////////////////////
{.$Define ReturnError}
UNIT TMTCD;
Interface
Type
  TMSF = Record
           Frm, Sec, Min, Trk : Byte;
         end;
  DiscRec = Record
    Length : TMSF;
    Track  : Array[1..99] of TMSF;
  end;

Const
  _CDDrivesNumber  : Integer = 0;                       {Number of CD drives}
  _CDCurrentDrv    : Integer = 0;                          {Current CD Drive}
  _CDFirstDrive    : Integer = 0;             {First CD drive, 0 for A: etc.}
  _MSCDEXVersion   : Word    = 0;        {High - major, Low  - minor version}
  {$IfDef ReturnError}
  _CDError         : Integer = 0;         {Procedures doesn't check this !!!}
  {$EndIf}
  {000Fh = invalid drive, 0001h = invalid function, 00FFh = no enough memory}

Var
  _MSCDEXInstalled : Boolean absolute _CDDrivesNumber;
  Request          : array[$00..$15] of Byte;
  TBuff            : array[$00..$0A] of Byte;

//////////////// TMSF Functions ////////////////
Function  HSG      (T:TMSF) : LongInt;
Function  RedBook(HSG :LongInt):TMSF;
Procedure HSG2RED  (HSG :LongInt; Var T:TMSF);
Procedure AddTMSF  (Var Dest : TMSF; Src : TMSF);
Procedure SubTMSF  (Var Dest : TMSF; Src : TMSF);
Function  EqTMSF   (T1, T2 : TMSF) : Boolean;
Function  NEqTMSF  (T1, T2 : TMSF) : Boolean;
Function  AddTMSFf (T1, T2 : TMSF) : TMSF;
Function  SubTMSFf (T1, T2 : TMSF) : TMSF;

OVERLOAD +   = AddTMSFf;
OVERLOAD -   = SubTMSFf;
OVERLOAD +:= = AddTMSF;
OVERLOAD -:= = SubTMSF;
OVERLOAD =   = EqTMSF;
OVERLOAD <>  = NEqTMSF;

///////////// Internal procedures //////////////
Procedure _CDInit;
Procedure _Stop;
Procedure PlayRequest;
Procedure InputRequest;
Procedure OutputRequest;
Procedure ReqQSubAudChnInfo;
Function  GetDriverName : String;
{$IfDef ReturnError}
Function  CDError       : Integer;
{$EndIf}

///////////////// CD commands //////////////////
Function  ChangeDrive (D:Word) : Boolean;
Procedure PauseAudio;
Procedure ResumeAudio;
Procedure CDStop;
Procedure EjectDoor;
Procedure InsertDoor;
Procedure LockDoor;
Procedure UnlockDoor;
Procedure SetVolume   (Vol    : Byte);
Procedure SeekPos     (SPos   : TMSF);
Procedure PlayRange   (From, Till   : TMSF);
Procedure PlayAmount  (From, Amount : TMSF);
Procedure PlayFrom    (From   : TMSF);
Procedure PlayTrack   (TrkNum : Byte);
Procedure PlaySingle  (TrkNum : Byte);
Procedure PlayReverse (Skip   : TMSF);
Procedure PlayForward (Skip   : TMSF);
Procedure PlayPrevTrk;
Procedure PlayNextTrk;
Procedure CDPrevTrack;
Procedure CDNextTrack;
Procedure ResetDrive;

/////////// Different status functions /////////
Function  AudVolume    : Byte;
Function  DeviceStatus : Word;
Function  MediaChanged : Byte;
Function  DATADisc     : Boolean;
Function  DoorOpen     : Boolean;
Function  Locked       : Boolean;
Function  NoDisc       : Boolean;
Function  DrvBusy      : Boolean;
Function  InPause      : Boolean;
Function  Writable     : Boolean;
Function  SuppAudVid   : Boolean;
Function  SuppAudChn   : Boolean;

////////////// Disc information ////////////////
Procedure GetDiscLen  (Var DLen : TMSF);
Procedure CurDiscPos  (Var DPos : TMSF);
Procedure CurTrkPos   (Var TPos : TMSF);
Procedure GetDiscRem  (Var DRem : TMSF);
Procedure GetTrkRem   (Var TRem : TMSF);
Procedure DiscTime    (Var Dest : TMSF);
Procedure TrkTime     (Var Dest : TMSF);
Procedure GetDiscInfo (Var Info : DiscRec);
Procedure GetTrkLoc   (Var TLoc : TMSF; Trk : Byte);  {Trk field = 1 if data}
Procedure GetTrkLen   (Var TLen : TMSF; Trk : Byte);  { and 0 if audio track}

Procedure AllStat1(Var _Open,_Locked,_NoDisc,_Busy,_Paused:Boolean);    {*}
Procedure AllStat2(Var DPos,TPos,DLen: TMSF; Var DATA:Boolean);         {*}

/////////////////////////////////////////////////////////////////////////////
{.$Define CorrectHSG} // HSG sector= minute*4500 + second*75 + frame - 150 //
/////////////////////////////////////////////////////////////////////////////
{.$Define DOS32Stub}                                 {Doesn't work, sorry...}
Implementation
Uses DPMI;
Var
  __HSG            : LongInt;
  __Trk            : Byte;
  PFromT, PCountT  : TMSF;
  Regs             : TRmRegs;
  RmSegment1,
  RmSegment2       : Word;
  OldExit          : Pointer;

Function HSG(T:TMSF) :LongInt;
  Result:=T.Min*4500+T.Sec*75+T.Frm-{$IfDef CorrectHSG}150{$Else}120{$EndIf};

Procedure HSG2RED(HSG :LongInt; Var T:TMSF);
  Begin
    HSG+:={$IfDef CorrectHSG}150{$Else}140{$EndIf};
    T.Frm:=HSG Mod 75;
    HSG:=HSG Div 75;
    T.Sec:=HSG Mod 60;
    T.Min:=HSG Div 60;
  End;

Function RedBook(HSG :LongInt):TMSF;
  Begin
    HSG+:={$IfDef CorrectHSG}150{$Else}140{$EndIf};
    Result.Frm:=HSG Mod 75;
    HSG:=HSG Div 75;
    Result.Sec:=HSG Mod 60;
    Result.Min:=HSG Div 60;
  End;

procedure AddTMSF(var Dest : TMSF; Src : TMSF);
Begin
  __HSG:=HSG(Dest)+HSG(Src);
  HSG2RED(__HSG,Dest);
End;

procedure SubTMSF(var Dest : TMSF; Src : TMSF);
Begin
  If HSG(Dest)>HSG(Src) Then
    __HSG:=HSG(Dest)-HSG(Src)
    Else __HSG:=HSG(Src)-HSG(Dest);
  HSG2RED(__HSG,Dest);
End;

function EqTMSF(T1, T2 : TMSF):Boolean;
  Result:=(T1.Trk=T2.Trk)and(T1.Min=T2.Min)and
           (T1.Sec=T2.Sec)and(T1.Frm=T2.Frm);

function NEqTMSF(T1, T2 : TMSF):Boolean;
  Result:=Not ((T1.Trk=T2.Trk)and(T1.Min=T2.Min)and
           (T1.Sec=T2.Sec)and(T1.Frm=T2.Frm));

function AddTMSFf(T1, T2 : TMSF):TMSF;
  Result:=RedBook(HSG(T1)+HSG(T2));

function SubTMSFf(T1, T2 : TMSF):TMSF;
  If HSG(T1)>HSG(T2) Then Result:=RedBook(HSG(T1)-HSG(T2))
                     Else Result:=RedBook(HSG(T2)-HSG(T1));

Procedure _CDInit;
Begin
  ClearRmRegs(Regs);
  Regs.AX:=$1500;
  RealModeInt($2F,Regs);
  _CDDrivesNumber:=Regs.bx;
  _CDFirstDrive:=Regs.cx;
  _CDCurrentDrv:=Regs.cx;
  ClearRmRegs(Regs);
  Regs.AX:=$150C;
  RealModeInt($2F,Regs);
  _MSCDEXVersion:=Regs.bx;
{$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
End;

procedure ReqQSubAudChnInfo; assembler;
asm
      mov     byte ptr TBuff,0Ch
      call    InputRequest
      mov     bh,byte ptr TBuff+2
      mov     cl,bh
      and     bh,0Fh
      shr     cl,4
      mov     al,10
      mul     cl
      add     bh,al
      mov     __Trk,bh
end;

procedure PlayRequest;
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$84;
  Mem[RmSegment1:$0D{$IfDef DOS32Stub}+_Zero{$EndIf}]:=0;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],Request,
       SizeOf(Request));
End;

procedure InputRequest;
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Move(TBuff,Mem[RmSegment2:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(TBuff));
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=3;
  MemW[RmSegment1:$0E{$IfDef DOS32Stub}+_Zero{$EndIf}]:=0;
  MemW[RmSegment1:$10{$IfDef DOS32Stub}+_Zero{$EndIf}]:=RmSegment2;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],Request,
       SizeOf(Request));
  Move(Mem[RmSegment2:0{$IfDef DOS32Stub}+_Zero{$EndIf}],TBuff,
       SizeOf(TBuff));
End;

procedure OutputRequest;
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Move(TBuff,Mem[RmSegment2:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(TBuff));
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$0C;
  MemW[RmSegment1:$0E{$IfDef DOS32Stub}+_Zero{$EndIf}]:=0;
  MemW[RmSegment1:$10{$IfDef DOS32Stub}+_Zero{$EndIf}]:=RmSegment2;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],Request,
       SizeOf(Request));
  Move(Mem[RmSegment2:0{$IfDef DOS32Stub}+_Zero{$EndIf}],TBuff,
       SizeOf(TBuff));
End;

procedure PauseAudio;
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$85;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],Request,
       SizeOf(Request));
End;

procedure ResumeAudio;
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$88;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],Request,
       SizeOf(Request));
End;

function DeviceStatus : Word; assembler;
asm
      mov     byte ptr TBuff,06h
      call    InputRequest
      mov     ax,word ptr TBuff+1
end;

procedure EjectDoor; assembler;
asm
      call    Locked      {Test if drive locked and exit procedure if it is.}
      cmp     al,0        {I add this stuff, because some drives still}
      jne     @Result     {ejecting, despite the fact that lock byte is on.}
      call    PauseAudio
      mov     byte ptr TBuff,00h
      call    OutputRequest
     @Result:
end;

function DoorOpen : Boolean; assembler;
asm                        {Some drives (like Pioneer, for example)}
      call    DeviceStatus {returns that there is no disc instead of}
      and     al,1 shl 0   {that door is opened.}
end;

function Locked : Boolean; assembler;
asm
      call    DeviceStatus
      not     al
      and     al,1 shl 1
end;

function NoDisc : Boolean; assembler;
asm
      call    DeviceStatus
      mov     al,ah
      and     al,1 shl 3
end;

function DrvBusy : Boolean; assembler;
asm
      call    DeviceStatus
      mov     al,byte ptr Request+4
      and     al,1 shl 1
end;

function InPause : Boolean; assembler;
asm
      mov     byte ptr TBuff,0Fh
      call    InputRequest
      mov     al,byte ptr TBuff+1
end;

function AudVolume : Byte; assembler;
asm
      mov     byte ptr TBuff,04h
      call    InputRequest
      mov     al,byte ptr TBuff+2
end;

function Writable : Boolean; assembler;
asm
      call    DeviceStatus
      and     al,1 shl 3
end;

function SuppAudVid : Boolean; assembler;
asm
      call    DeviceStatus
      and     al,1 shl 4
end;

function SuppAudChn : Boolean; assembler;
asm
      call    DeviceStatus
      mov     al,ah
      and     al,1 shl 0
end;

procedure InsertDoor; assembler;
asm
      mov     byte ptr TBuff,05h
      call    OutputRequest
end;

procedure LockDoor; assembler;
asm
      mov     word ptr TBuff,0101h
      call    OutputRequest
end;

procedure UnlockDoor; assembler;
asm
      mov     word ptr TBuff,0001h
      call    OutputRequest
end;

procedure SetVolume(Vol : Byte); assembler;
asm
      mov     byte ptr TBuff,03h
      xor     al,al
      mov     ah,Vol
      mov     word ptr TBuff+1,ax
      inc     al
      mov     word ptr TBuff+3,ax
      inc     al
      mov     word ptr TBuff+5,ax
      inc     al
      mov     word ptr TBuff+7,ax
      call    OutputRequest
end;

procedure GetDiscLen(var DLen : TMSF); Assembler;
asm
      mov     byte ptr TBuff,0Ah
      call    InputRequest
      mov     edi,[DLen]
      mov     ah,byte ptr TBuff+2
      mov     [edi+3],ah
      mov     ah,byte ptr TBuff+3
      mov     [edi],ah
      mov     ah,byte ptr TBuff+4
      mov     [edi+1],ah
      mov     ah,byte ptr TBuff+5
      mov     [edi+2],ah
end;

procedure GetTrkLoc(var TLoc : TMSF; Trk : Byte); Assembler;
asm
      mov     byte ptr TBuff,0Bh
      mov     al,Trk
      mov     byte ptr TBuff+1,al
      call    InputRequest
      mov     edi,[TLoc]
      mov     ah,byte ptr TBuff+2
      mov     [edi],ah
      mov     ah,byte ptr TBuff+3
      mov     [edi+1],ah
      mov     ah,byte ptr TBuff+4
      mov     [edi+2],ah
      mov     ah,byte ptr TBuff+6
      and     ah,1 shl 6
      shr     ah,6
      mov     [edi+3],ah
End;

procedure GetTrkLen(var TLen : TMSF; Trk : Byte);
var
  TLoc : TMSF;
begin
  GetTrkLoc(TLen, Trk);
  GetDiscLen(TLoc);
  if Trk < TLoc.Trk then
    GetTrkLoc(TLoc, Trk+1);
  SubTMSF(TLen, TLoc);
end;


procedure GetDiscInfo(var Info : DiscRec);
var
  Trk : Byte;
begin
  GetDiscLen(Info.Length);
  for Trk := 1 to Info.Length.Trk do
    GetTrkLen(Info.Track[Trk], Trk);
end;

procedure CurDiscPos(var DPos : TMSF ); Assembler;
asm
      call    ReqQSubAudChnInfo
      mov     edi,[DPos]
      mov     ah,__Trk
      mov     [edi+3],ah
      mov     ah,byte ptr TBuff+8
      mov     [edi+2],ah
      mov     ah,byte ptr TBuff+9
      mov     [edi+1],ah
      mov     ah,byte ptr TBuff+0Ah
      mov     [edi],ah
end;

procedure CurTrkPos(var TPos : TMSF ); Assembler;
asm
      call    ReqQSubAudChnInfo
      mov     edi,[TPos]
      mov     ah,__Trk
      mov     [edi+3],ah
      mov     ah,byte ptr TBuff+4
      mov     [edi+2],ah
      mov     ah,byte ptr TBuff+5
      mov     [edi+1],ah
      mov     ah,byte ptr TBuff+6
      mov     [edi],ah
End;

procedure GetDiscRem(var DRem : TMSF);
var
  DLen : TMSF;
begin
  GetDiscLen(DLen);
  CurDiscPos(DRem);
  SubTMSF(DRem, DLen);
end;

procedure GetTrkRem(var TRem : TMSF);
var
  TLen : TMSF;
begin
  GetTrkLen(TLen, TRem.Trk);
  CurTrkPos(TRem);
  SubTMSF(TRem, TLen);
end;

procedure DiscTime(var Dest : TMSF);
var
  TLoc : TMSF;
begin
  GetTrkLoc(TLoc, Dest.Trk);
  AddTMSF(Dest, TLoc);
  Dest.Trk := 0;
end;

procedure TrkTime(var Dest : TMSF);
var
  Upper, Lower, Middle : Byte;
  Disc, Trk : TMSF;
begin
  Dest.Trk := 0;
  GetDiscLen(Disc);
  GetTrkLoc(Trk, Disc.Trk);
  if LongInt(Dest) >= (LongInt(Trk) and $FFFFFF) then
    Dest.Trk := Disc.Trk;
  if (Disc.Trk > 1) and (Dest.Trk = 0) then
  begin
    GetTrkLoc(Trk, 2);
    if Longint(Dest) < (Longint(Trk) and $FFFFFF) then
    begin
      Dest.Trk := 1;
      GetTrkLoc(Trk, 1);
    end;
  end;
  if Dest.Trk = 0 then
  begin
    Lower := 2;
    Upper := Disc.Trk - 1;
    while Dest.Trk <> Middle do
    begin
      Middle := (Lower + Upper) shr 1;
      GetTrkLoc(Trk, Middle + 1);
      if Longint(Dest) > (Longint(Trk) and $FFFFFF) then
      begin
        Lower := Middle + 1;
        Continue;
      end;
      GetTrkLoc(Trk, Middle);
      if Longint(Dest) < (Longint(Trk) and $FFFFFF) then
        Upper := Middle - 1
      else
        Dest.Trk := Middle;
    end;
  end;
  SubTMSF(Dest, Trk);
end;

procedure PlayFrom(From : TMSF);
begin
  PauseAudio;
  GetDiscLen(PCountT);
  if From.Trk <> 0 then DiscTime(From);
  Move(HSG(From),Request[$0E],4);
  Move(HSG(PCountT) - HSG(From),Request[$12],4);
  PlayRequest;
end;

procedure PlayRange(From, Till : TMSF);
begin
  PauseAudio;
  if From.Trk <> 0 then DiscTime(From);
  if Till.Trk <> 0 then DiscTime(Till);
  Move(HSG(From),Request[$0E],4);
  Move(HSG(PCountT) - HSG(From),Request[$12],4);
  PlayRequest;
end;

procedure PlayAmount(From, Amount : TMSF);
begin
  PauseAudio;
  if From.Trk <> 0 then DiscTime(From);
  Move(HSG(From),Request[$0E],4);
  Move(HSG(Amount),Request[$12],4);
  PlayRequest;
end;

procedure PlayTrack(TrkNum : Byte);
begin
  GetTrkLoc(PFromT, TrkNum);
  if PFromT.Trk = 0 then PlayFrom(PFromT);
end;

procedure PlaySingle(TrkNum : Byte);
begin
  GetTrkLoc(PFromT, TrkNum);
  GetTrkLen(PCountT, TrkNum);
  if PFromT.Trk = 0 then PlayAmount(PFromT, PCountT);
end;

procedure PlayPrevTrk;
Var PCount:TMSF;
begin
  GetDiscLen(PCount);
  CurTrkPos(PFromT);
  if PFromT.Trk = 1 then
    PFromT.Trk := PCountT.Trk
  else
    Dec(PFromT.Trk);
  GetTrkLoc(PFromT, PFromT.Trk);
  if PFromT.Trk = 0 then PlayFrom(PFromT);
end;

procedure PlayNextTrk;
begin
  GetDiscLen(PCountT);
  CurTrkPos(PFromT);
  if PFromT.Trk = PCountT.Trk then
    PFromT.Trk := 1
  else
    Inc(PFromT.Trk);
  GetTrkLoc(PFromT, PFromT.Trk);
  if PFromT.Trk = 0 then PlayFrom(PFromT);
end;

procedure PlayReverse(Skip : TMSF);
begin
  CurDiscPos(PFromT);
  SubTMSF(PFromT, Skip);
  PFromT.Trk := 0;
  PlayFrom(PFromT);
end;

procedure PlayForward(Skip : TMSF);
begin
  CurDiscPos(PFromT);
  AddTMSF(PFromT, Skip);
  PFromT.Trk := 0;
  PlayFrom(PFromT);
end;

Function DATADisc:Boolean; Assembler;
Asm
 mov     byte ptr TBuff,0Ah           {Checking how many tracks}
 call    InputRequest                 {are on current disk.}
 mov     ah,byte ptr TBuff+2          {If disk have only one}
 xor     al,al                        {track, we checking if it's}
 cmp     ah,1                         {DATA track. Otherwise it means that}
 jne     @Return                      {disc audio, or combined audio/data.}
 mov     byte ptr TBuff,0Bh
 mov     byte ptr TBuff+1,1
 call    InputRequest
 mov     al,byte ptr TBuff+6
 and     al,1 shl 6
 shr     al,6
@Return:
End;

procedure ResetDrive; assembler;
asm
      mov     byte ptr TBuff,02h  {Warning !!! A lot of drivers process}
      call    OutputRequest       {this procedure incorrectly !!!}
end;

Function MediaChanged : Byte; Assembler;
{  1  :  Media not changed            }
{  0  :  Don't Know                   }
{ -1  :  Media changed.               }
{ N.B. Usually works incorrectly ;-)  }
asm
      mov     byte ptr TBuff,09h
      call    OutputRequest
      mov     al,byte ptr TBuff+1
end;

Function GetDriverName : String;     {I never use this procedure, but}
Var                                  {if you want to work with MSCDEX}
   Cdntemp : Array[1..18] Of Byte;   {through Int 21h it will be usefull}
   Count : Byte;
   Cdstemp : String[8];
Begin
  ClearRmRegs(Regs);
  Move(Request,Mem[RmSegment1:0{$IfDef DOS32Stub}+_Zero{$EndIf}],
       SizeOf(Request));
  Regs.AX:=$1501;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$84;
  Mem[RmSegment1:$0D{$IfDef DOS32Stub}+_Zero{$EndIf}]:=0;
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
  Move(Mem[MemW[RmSegment1:3{$IfDef DOS32Stub}+_Zero{$EndIf}]:
           MemW[RmSegment1:1{$IfDef DOS32Stub}+_Zero{$EndIf}]
           {$IfDef DOS32Stub}+_Zero{$EndIf}],CDNTemp,18);
  Count := 1;
  Repeat
     CDStemp[Count] := CHR(CDNTemp[10+Count]);
     Inc(Count);
  Until (Count > 8) Or (Cdntemp[10+Count]=32);
  CDSTemp[0] := CHR(Count-1);
  GetDriverName := CDSTemp;
End;

Procedure CDStop;
Var T:TMSF;
Begin
  CurDiscPos(T);              {This stuff is necessary because, some drivers}
  GetTrkLoc(T,T.Trk);         {doesn't process stop (double pause) procedure}
  SeekPos(T);                 {correctly}
  PauseAudio;
  If InPause Then PauseAudio;
End;

{Procedure for my CD player.}
Procedure AllStat1(Var _Open,_Locked,_NoDisc,_Busy,_Paused:Boolean); Assembler;
asm
      call    DeviceStatus
      mov     bl,al
      and     al,1
      mov     edi,[_Open]
      mov     [edi],al
      mov     al,bl
      not     al
      and     al,1 shl 1
      mov     edi,[_Locked]
      mov     [edi],al
      mov     al,ah
      and     al,1 shl 3
      mov     edi,[_NoDisc]
      mov     [edi],al
      mov     al,byte ptr Request+4
      and     al,1 shl 1
      mov     edi,[_Busy]
      mov     [edi],al
      mov     byte ptr TBuff,0Fh
      call    InputRequest
      mov     al,byte ptr TBuff+1
      mov     edi,[_Paused]
      mov     [edi],al
end;

{Procedure for my CD player.}
Procedure AllStat2(Var DPos,TPos,DLen: TMSF; Var DATA:Boolean); Assembler;
asm
      call    ReqQSubAudChnInfo
      mov     edi,[DPos]
      mov     bl,byte ptr TBuff+8
      mov     [edi+2],bl
      mov     ah,byte ptr TBuff+9
      mov     [edi+1],ah
      mov     al,byte ptr TBuff+0Ah
      mov     [edi],al
      mov     edi,[TPos]
      mov     bl,byte ptr TBuff+4
      mov     [edi+2],bl
      mov     ah,byte ptr TBuff+5
      mov     [edi+1],ah
      mov     al,byte ptr TBuff+6
      mov     [edi],al
      mov     byte ptr TBuff,0Ah
      call    InputRequest
      mov     edi,[DLen]
      mov     ah,byte ptr TBuff+2
      mov     [edi+3],ah
      mov     ah,byte ptr TBuff+3
      mov     [edi],ah
      mov     ah,byte ptr TBuff+4
      mov     [edi+1],ah
      mov     ah,byte ptr TBuff+5
      mov     [edi+2],ah
      mov     byte ptr TBuff,0Bh
      mov     al,__Trk
      mov     byte ptr TBuff+1,al
      call    InputRequest
      mov     al,byte ptr TBuff+6
      and     al,1 shl 6
      shr     al,6
      mov     edi,[Data]
      mov     [edi],al
      mov     al,__Trk
      mov     edi,[DPos]
      mov     [edi+3],al
      mov     edi,[TPos]
      mov     [edi+3],al
end;

Function  ChangeDrive (D:Word):Boolean; {0=A:,1=B: etc.}
Begin
  ClearRmRegs(Regs);
  Regs.AX:=$150B;
  Regs.CX:=D;
  RealModeInt($2F,Regs);
  If Regs.AX <> 0 Then
   Begin
    _CDCurrentDrv:=D;
    ChangeDrive:=True;
   End Else ChangeDrive:=False;
End;
{$IfDef ReturnError}
Function CDError:Integer; Assembler;
Asm
 Mov AX,_CDError
 Mov _CDError,0
End;
{$EndIf}
procedure SeekPos(SPos : TMSF);
Begin
  PauseAudio;
  if SPos.Trk <> 0 then GetTrkLoc(SPos, SPos.Trk);
  ClearRmRegs(Regs);
  Regs.AX:=$1510;
  Regs.ES:=RmSegment1;
  Regs.CX:=_CDCurrentDrv;
  Mem[RmSegment1:2{$IfDef DOS32Stub}+_Zero{$EndIf}]:=$83;
  Mem[RmSegment1:$0D{$IfDef DOS32Stub}+_Zero{$EndIf}]:=0;
  Move(HSG(SPos),Mem[RmSegment1:$14{$IfDef DOS32Stub}+_Zero{$EndIf}],4);
  RealModeInt($2F,Regs);
  {$IfDef ReturnError}If (Regs.Flags and 1)=1 Then _CDError:=Regs.AX;{$EndIf}
end;

procedure CDPrevTrack;
Var PCount:TMSF;
begin
  CurTrkPos(PFromT);
  if PFromT.Trk = 1 then
  begin
    GetDiscLen(PCount);
    Move(PCount,Request[$12],4);
    PFromT.Trk := PCountT.Trk;
  end
  else
    Dec(PFromT.Trk);
  GetTrkLoc(PFromT, PFromT.Trk);
  if PFromT.Trk = 0 then
   begin
    If DrvBusy Then PlayFrom(PFromT)
               Else SeekPos(PFromT)
   end;
end;

procedure CDNextTrack;
begin
  GetDiscLen(PCountT);
  CurTrkPos(PFromT);
  if PFromT.Trk = PCountT.Trk then
    PFromT.Trk := 1
  else
    Inc(PFromT.Trk);
  GetTrkLoc(PFromT, PFromT.Trk);
  if PFromT.Trk = 0 then
   begin
    If DrvBusy Then PlayFrom(PFromT)
               Else SeekPos(PFromT)
   end;
end;


/////////////////////////////////////////////////////////////////////////////
Procedure _Stop;
Begin
 If RmSegment1<>0 Then DOSMemoryFree(RmSegment1);
 If RmSegment2<>0 Then DOSMemoryFree(RmSegment2);
 ExitProc:=OldExit;
End;
Begin
  _cdinit;
  RmSegment1:=DOSMemoryAlloc(SizeOf(Request));
  RmSegment2:=DOSMemoryAlloc(SizeOf(TBuff));
  If (RmSegment1=0)or(RmSegment2=0) Then
  {$IfDef ReturnError}
  _CDError:=$FF;
  {$Else}
  Begin
   Writeln('No enougth free DOS memory.');
   Halt($CF)
  End;
  {$EndIf}
  OldExit:=ExitProc;
  ExitProc:=@_Stop;
End.